/* Native-dependent code for GNU/Linux x86-64.

   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006
   Free Software Foundation, Inc.
   Contributed by Jiri Smid, SuSE Labs.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "inferior.h"
#include "gdbcore.h"
#include "regcache.h"
#include "observer.h"
#include "linux-nat.h"

#include "gdb_assert.h"
#include "gdb_string.h"
#include <sys/ptrace.h>
#include <sys/debugreg.h>
#include <sys/syscall.h>
#include <sys/procfs.h>
#include <asm/prctl.h>
/* FIXME ezannoni-2003-07-09: we need <sys/reg.h> to be included after
   <asm/ptrace.h> because the latter redefines FS and GS for no apparent
   reason, and those definitions don't match the ones that libpthread_db
   uses, which come from <sys/reg.h>.  */
/* ezannoni-2003-07-09: I think this is fixed. The extraneous defs have
   been removed from ptrace.h in the kernel.  However, better safe than
   sorry.  */
#include <asm/ptrace.h>
#include <sys/reg.h>
#include "gdb_proc_service.h"

/* Prototypes for supply_gregset etc.  */
#include "gregset.h"

#include "amd64-tdep.h"
#include "i386-linux-tdep.h"
#include "amd64-nat.h"
#include "i387-tdep.h"
#include "bfd/elf-bfd.h"
#include "gdb_procfs32.h"

/* Mapping between the general-purpose registers in GNU/Linux x86-64
   `struct user' format and GDB's register cache layout.  */

static int amd64_linux_gregset64_reg_offset[] =
{
  RAX * 8, RBX * 8,		/* %rax, %rbx */
  RCX * 8, RDX * 8,		/* %rcx, %rdx */
  RSI * 8, RDI * 8,		/* %rsi, %rdi */
  RBP * 8, RSP * 8,		/* %rbp, %rsp */
  R8 * 8, R9 * 8,		/* %r8 ... */
  R10 * 8, R11 * 8,
  R12 * 8, R13 * 8,
  R14 * 8, R15 * 8,		/* ... %r15 */
  RIP * 8, EFLAGS * 8,		/* %rip, %eflags */
  CS * 8, SS * 8,		/* %cs, %ss */
  DS * 8, ES * 8,		/* %ds, %es */
  FS * 8, GS * 8		/* %fs, %gs */
};


/* Mapping between the general-purpose registers in GNU/Linux x86-64
   `struct user' format and GDB's register cache layout for GNU/Linux
   i386.

   Note that most GNU/Linux x86-64 registers are 64-bit, while the
   GNU/Linux i386 registers are all 32-bit, but since we're
   little-endian we get away with that.  */

/* This info is not reusable from "i386-linux-nat.c" as gdb itself runs in
   64-bit mode and so ptrace(2) has 64-bit structure layout.
   Just the corefile being generated has 32-bit layout so we need to do
   a conversion specific to the i386-on-amd64 compatibility mode.  */
static int i386_linux_gregset32_reg_offset[] =
{
  6 * 4,			/* %eax */
  1 * 4,			/* %ecx */
  2 * 4,			/* %edx */
  0 * 4,			/* %ebx */
  15 * 4,			/* %esp */
  5 * 4,			/* %ebp */
  3 * 4,			/* %esi */
  4 * 4,			/* %edi */
  12 * 4,			/* %eip */
  14 * 4,			/* %eflags */
  13 * 4,			/* %cs */
  16 * 4,			/* %ss */
  7 * 4,			/* %ds */
  8 * 4,			/* %es */
  9 * 4,			/* %fs */
  10 * 4,			/* %gs */
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1,
  11 * 4			/* "orig_eax" */
};

/* From <sys/reg.h> on GNU/Linux i386.  */
static int amd64_linux_gregset32_reg_offset[] =
{
  RAX * 8, RCX * 8,		/* %eax, %ecx */
  RDX * 8, RBX * 8,		/* %edx, %ebx */
  RSP * 8, RBP * 8,		/* %esp, %ebp */
  RSI * 8, RDI * 8,		/* %esi, %edi */
  RIP * 8, EFLAGS * 8,		/* %eip, %eflags */
  CS * 8, SS * 8,		/* %cs, %ss */
  DS * 8, ES * 8,		/* %ds, %es */
  FS * 8, GS * 8,		/* %fs, %gs */
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1,
  ORIG_RAX * 8			/* "orig_eax" */
};


/* This functions make ELF32 32-bit elfcore note sections
   on amd64 environment. */

static char *
amd64_linux_elfcore_write_prpsinfo (bfd *abfd, char *buf, int *bufsiz,
				    const char *fname, const char *psargs)
{
  if (gdbarch_ptr_bit(current_gdbarch) == 32)
    {
      int note_type;
      char *note_name = "CORE";
      struct elf_prpsinfo32 data;
      note_type = NT_PRPSINFO;

      memset (&data, 0, sizeof (data));
      strncpy (data.pr_fname, fname, sizeof (data.pr_fname));
      strncpy (data.pr_psargs, psargs, sizeof (data.pr_psargs));
      return elfcore_write_note (abfd, buf, bufsiz,
				 note_name, note_type, &data, sizeof (data));
    }
  else
    return elfcore_write_prpsinfo (abfd, buf, bufsiz, fname, psargs);
}

static void
amd64_linux_set_registers (const gdb_byte *gregs, gdb_byte *buf)
{
  int i;
  /* Copy the i386 registers in the amd64 layout into i386 layout.  */
  for (i = 0; i < I386_NUM_GREGS; i++)
    memcpy(buf + i386_linux_gregset32_reg_offset[i],
	   gregs + amd64_linux_gregset32_reg_offset[i], 4);
  for (i = I386_CS_REGNUM; i <= I386_GS_REGNUM; i++)
    memcpy(buf + i386_linux_gregset32_reg_offset[i],
	   gregs + amd64_linux_gregset32_reg_offset[i], 4);
}

static char *
amd64_linux_elfcore_write_prstatus (bfd *abfd, char *buf, int *bufsiz,
			      long pid, int cursig, const void *gregs)
{
  if (gdbarch_ptr_bit(current_gdbarch) == 32)
    {
      char *note_name = "CORE";
      struct elf_prstatus32 prstat;
      memset (&prstat, 0, sizeof (prstat));
      prstat.pr_pid = pid;
      prstat.pr_cursig = cursig;
      amd64_linux_set_registers (gregs, (gdb_byte *) &prstat.pr_reg);
      return elfcore_write_note (abfd, buf, bufsiz, note_name,
				 NT_PRSTATUS, &prstat, sizeof (prstat));
    }
  else
    return elfcore_write_prstatus (abfd, buf, bufsiz, pid, cursig, gregs);
}

static char *
amd64_elfcore_write_prxfpreg32 (bfd *abfd, char *buf, int *bufsiz)
{
  char *note_name = "LINUX";
  elf_fpxregset32_t fpxregs32;

  i387_fill_fxsave (&fpxregs32, -1);
  return elfcore_write_note(abfd, buf, bufsiz,
				     note_name, NT_PRXFPREG, &fpxregs32,
				     sizeof(fpxregs32));
}

static char *
amd64_linux_elfcore_write_prfpreg (bfd *abfd, char *buf,
			     int *bufsiz, const void *fpregs, int size)
{
  if (gdbarch_ptr_bit(current_gdbarch) == 32)
    {
      char *note_name = "CORE";
      elf_fpregset32_t fpregs32;
  
      i387_fill_fsave (&fpregs32, -1);
      buf = elfcore_write_note(abfd, buf, bufsiz, note_name,
			       NT_FPREGSET, &fpregs32, sizeof(fpregs32));

      return amd64_elfcore_write_prxfpreg32 (abfd, buf, bufsiz);
    }
  else
    return elfcore_write_prfpreg (abfd, buf, bufsiz, fpregs, size);
}


/* Transfering the general-purpose registers between GDB, inferiors
   and core files.  */

/* Fill GDB's register cache with the general-purpose register values
   in *GREGSETP.  */

void
supply_gregset (elf_gregset_t *gregsetp)
{
  amd64_supply_native_gregset (current_regcache, gregsetp, -1);
}

/* Fill register REGNUM (if it is a general-purpose register) in
   *GREGSETP with the value in GDB's register cache.  If REGNUM is -1,
   do this for all registers.  */

void
fill_gregset (elf_gregset_t *gregsetp, int regnum)
{
  amd64_collect_native_gregset (current_regcache, gregsetp, regnum);
}

/* Transfering floating-point registers between GDB, inferiors and cores.  */

/* Fill GDB's register cache with the floating-point and SSE register
   values in *FPREGSETP.  */

void
supply_fpregset (elf_fpregset_t *fpregsetp)
{
  amd64_supply_fxsave (current_regcache, -1, fpregsetp);
}

/* Fill register REGNUM (if it is a floating-point or SSE register) in
   *FPREGSETP with the value in GDB's register cache.  If REGNUM is
   -1, do this for all registers.  */

void
fill_fpregset (elf_fpregset_t *fpregsetp, int regnum)
{
  amd64_collect_fxsave (current_regcache, regnum, fpregsetp);
}


/* Transferring arbitrary registers between GDB and inferior.  */

/* Fetch register REGNUM from the child process.  If REGNUM is -1, do
   this for all registers (including the floating point and SSE
   registers).  */

static void
amd64_linux_fetch_inferior_registers (int regnum)
{
  int tid;

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  if (regnum == -1 || amd64_native_gregset_supplies_p (regnum))
    {
      elf_gregset_t regs;

      if (ptrace (PTRACE_GETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't get registers"));

      amd64_supply_native_gregset (current_regcache, &regs, -1);
      if (regnum != -1)
	return;
    }

  if (regnum == -1 || !amd64_native_gregset_supplies_p (regnum))
    {
      elf_fpregset_t fpregs;

      if (ptrace (PTRACE_GETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't get floating point status"));

      amd64_supply_fxsave (current_regcache, -1, &fpregs);
    }
}

/* Store register REGNUM back into the child process.  If REGNUM is
   -1, do this for all registers (including the floating-point and SSE
   registers).  */

static void
amd64_linux_store_inferior_registers (int regnum)
{
  int tid;

  /* GNU/Linux LWP ID's are process ID's.  */
  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid); /* Not a threaded program.  */

  if (regnum == -1 || amd64_native_gregset_supplies_p (regnum))
    {
      elf_gregset_t regs;

      if (ptrace (PTRACE_GETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't get registers"));

      amd64_collect_native_gregset (current_regcache, &regs, regnum);

      if (ptrace (PTRACE_SETREGS, tid, 0, (long) &regs) < 0)
	perror_with_name (_("Couldn't write registers"));

      if (regnum != -1)
	return;
    }

  if (regnum == -1 || !amd64_native_gregset_supplies_p (regnum))
    {
      elf_fpregset_t fpregs;

      if (ptrace (PTRACE_GETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't get floating point status"));

      amd64_collect_fxsave (current_regcache, regnum, &fpregs);

      if (ptrace (PTRACE_SETFPREGS, tid, 0, (long) &fpregs) < 0)
	perror_with_name (_("Couldn't write floating point status"));

      return;
    }
}


static unsigned long
amd64_linux_dr_get (ptid_t ptid, int regnum)
{
  int tid;
  unsigned long value;

  tid = TIDGET (ptid);
  if (tid == 0)
    tid = PIDGET (ptid);

  /* FIXME: kettenis/2001-03-27: Calling perror_with_name if the
     ptrace call fails breaks debugging remote targets.  The correct
     way to fix this is to add the hardware breakpoint and watchpoint
     stuff to the target vectore.  For now, just return zero if the
     ptrace call fails.  */
  errno = 0;
  value = ptrace (PT_READ_U, tid,
		  offsetof (struct user, u_debugreg[regnum]), 0);
  if (errno != 0)
#if 0
    perror_with_name (_("Couldn't read debug register"));
#else
    return 0;
#endif

  return value;
}

static void
amd64_linux_dr_set (ptid_t ptid, int regnum, unsigned long value)
{
  int tid;

  tid = TIDGET (ptid);
  if (tid == 0)
    tid = PIDGET (ptid);

  errno = 0;
  ptrace (PT_WRITE_U, tid, offsetof (struct user, u_debugreg[regnum]), value);
  if (errno != 0)
    perror_with_name (_("Couldn't write debug register"));
}

void
amd64_linux_dr_set_control (ptid_t ptid, unsigned long control)
{
  amd64_linux_dr_set (ptid, DR_CONTROL, control);
}

void
amd64_linux_dr_set_addr (ptid_t ptid, int regnum, CORE_ADDR addr)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  amd64_linux_dr_set (ptid, DR_FIRSTADDR + regnum, addr);
}

void
amd64_linux_dr_reset_addr (ptid_t ptid, int regnum)
{
  gdb_assert (regnum >= 0 && regnum <= DR_LASTADDR - DR_FIRSTADDR);

  amd64_linux_dr_set (ptid, DR_FIRSTADDR + regnum, 0L);
}

unsigned long
amd64_linux_dr_get_status (ptid_t ptid)
{
  return amd64_linux_dr_get (ptid, DR_STATUS);
}


/* Structure used to sync debug registers for all threads.  */
struct amd64_debug_register_state
{
  int tid;
  CORE_ADDR addr[DR_LASTADDR - DR_FIRSTADDR + 1];
  unsigned long control;
};

static void
amd64_linux_set_debug_regs_for_thread (ptid_t ptid,
				       struct amd64_debug_register_state *dbs)
{
  int i;
  for (i = 0; i < (DR_LASTADDR - DR_FIRSTADDR) + 1; ++i)
    amd64_linux_dr_set_addr (ptid, i, dbs->addr[i]);
  amd64_linux_dr_set_control (ptid, dbs->control);
}

/* Iterator function to support syncing debug registers across all threads.  */
static int
amd64_linux_sync_debug_registers_callback (struct lwp_info *lwp, void *data)
{
  struct amd64_debug_register_state *args = data;
  int i, tid;

  tid = TIDGET (lwp->ptid);
  if (tid == 0)
    tid = PIDGET (lwp->ptid);

  if (tid != args->tid)
    amd64_linux_set_debug_regs_for_thread (lwp->ptid, args);
  return 0;
}

/* Sync the debug registers for all known threads to the current
   thread that has just performed an operation.  This is required
   because the debug registers are thread-specific.  We want
   watchpoints and hardware breakpoints to be treated globally
   across all threads.  */
static int
amd64_linux_sync_debug_registers_across_threads (void)
{
  int i, tid;
  struct amd64_debug_register_state args;

  tid = TIDGET (inferior_ptid);
  if (tid == 0)
    tid = PIDGET (inferior_ptid);

  args.tid = tid;
  for (i = 0; i < (DR_LASTADDR - DR_FIRSTADDR) + 1; ++i)
    args.addr[i] = amd64_linux_dr_get (inferior_ptid, DR_FIRSTADDR + i);
  args.control = amd64_linux_dr_get (inferior_ptid, DR_CONTROL);

  iterate_over_lwps (&amd64_linux_sync_debug_registers_callback, &args);

  return 0;
}

/* Insert a watchpoint to watch a memory region which starts at
   address ADDR and whose length is LEN bytes.  Watch memory accesses
   of the type TYPE.  Return 0 on success, -1 on failure.  */
int
amd64_linux_insert_watchpoint (CORE_ADDR addr, int len, int type)
{
  int rc;
  rc = i386_insert_watchpoint (addr, len, type);
  if (!rc)
    amd64_linux_sync_debug_registers_across_threads ();
  return rc;
}

/* Remove a watchpoint that watched the memory region which starts at
   address ADDR, whose length is LEN bytes, and for accesses of the
   type TYPE.  Return 0 on success, -1 on failure.  */
int
amd64_linux_remove_watchpoint (CORE_ADDR addr, int len, int type)
{
  int rc;
  rc = i386_remove_watchpoint (addr, len, type);
  if (!rc)
    amd64_linux_sync_debug_registers_across_threads ();
  return rc;
}

/* Insert a hardware-assisted breakpoint at address ADDR.  SHADOW is
   unused.  Return 0 on success, EBUSY on failure.  */
int
amd64_linux_insert_hw_breakpoint (struct bp_target_info *bp_tgt)
{
  int rc;
  rc = i386_insert_hw_breakpoint (bp_tgt);
  if (!rc)
    amd64_linux_sync_debug_registers_across_threads ();
  return rc;
}

/* Remove a hardware-assisted breakpoint at address ADDR.  SHADOW is
   unused.  Return 0 on success, -1 on failure.  */
int
amd64_linux_remove_hw_breakpoint (struct bp_target_info *bp_tgt)
{
  int rc;
  rc = i386_remove_hw_breakpoint (bp_tgt);
  if (!rc)
    amd64_linux_sync_debug_registers_across_threads ();
  return rc;
}

/* Observer function for a new thread attach.  We need to insert
   existing watchpoints and hardware breakpoints on the new thread.  */
static void
amd64_linux_new_thread (ptid_t ptid)
{
  int i;
  struct amd64_debug_register_state dbs;

  for (i = 0; i < (DR_LASTADDR - DR_FIRSTADDR) + 1; ++i)
    dbs.addr[i] = amd64_linux_dr_get (inferior_ptid, DR_FIRSTADDR + i);
  dbs.control = amd64_linux_dr_get (inferior_ptid, DR_CONTROL);

  amd64_linux_set_debug_regs_for_thread (ptid, &dbs);
}

/* This function is called by libthread_db as part of its handling of
   a request for a thread's local storage address.  */

ps_err_e
ps_get_thread_area (const struct ps_prochandle *ph,
                    lwpid_t lwpid, int idx, void **base)
{
  if (gdbarch_ptr_bit (current_gdbarch) == 32)
    {
      /* The full structure is found in <asm-i386/ldt.h>.  The second
	 integer is the LDT's base_address and that is used to locate
	 the thread's local storage.  See i386-linux-nat.c more
	 info.  */
      unsigned int desc[4];

      /* This code assumes that "int" is 32 bits and that
	 GET_THREAD_AREA returns no more than 4 int values.  */
      gdb_assert (sizeof (int) == 4);	
#ifndef PTRACE_GET_THREAD_AREA
#define PTRACE_GET_THREAD_AREA 25
#endif
      if  (ptrace (PTRACE_GET_THREAD_AREA, 
		   lwpid, (void *) (long) idx, (unsigned long) &desc) < 0)
	return PS_ERR;
      
      /* Extend the value to 64 bits.  Here it's assumed that a "long"
	 and a "void *" are the same.  */
      (*base) = (void *) (long) desc[1];
      return PS_OK;
    }
  else
    {
      /* This definition comes from prctl.h, but some kernels may not
         have it.  */
#ifndef PTRACE_ARCH_PRCTL
#define PTRACE_ARCH_PRCTL      30
#endif
      /* FIXME: ezannoni-2003-07-09 see comment above about include
	 file order.  We could be getting bogus values for these two.  */
      gdb_assert (FS < ELF_NGREG);
      gdb_assert (GS < ELF_NGREG);
      switch (idx)
	{
	case FS:
	  if (ptrace (PTRACE_ARCH_PRCTL, lwpid, base, ARCH_GET_FS) == 0)
	    return PS_OK;
	  break;
	case GS:
	  if (ptrace (PTRACE_ARCH_PRCTL, lwpid, base, ARCH_GET_GS) == 0)
	    return PS_OK;
	  break;
	default:                   /* Should not happen.  */
	  return PS_BADADDR;
	}
    }
  return PS_ERR;               /* ptrace failed.  */
}


static void (*super_post_startup_inferior) (ptid_t ptid);

static void
amd64_linux_child_post_startup_inferior (ptid_t ptid)
{
  i386_cleanup_dregs ();
  super_post_startup_inferior (ptid);
}


/* Provide a prototype to silence -Wmissing-prototypes.  */
void _initialize_amd64_linux_nat (void);

void
_initialize_amd64_linux_nat (void)
{
  struct target_ops *t;

  amd64_native_gregset32_reg_offset = amd64_linux_gregset32_reg_offset;
  amd64_native_gregset32_num_regs = I386_LINUX_NUM_REGS;
  amd64_native_gregset64_reg_offset = amd64_linux_gregset64_reg_offset;

  gdb_assert (ARRAY_SIZE (amd64_linux_gregset32_reg_offset)
	      == amd64_native_gregset32_num_regs);
  gdb_assert (ARRAY_SIZE (amd64_linux_gregset64_reg_offset)
	      == amd64_native_gregset64_num_regs);

  /* Fill in the generic GNU/Linux methods.  */
  t = linux_target ();

  /* Override the GNU/Linux inferior startup hook.  */
  super_post_startup_inferior = t->to_post_startup_inferior;
  t->to_post_startup_inferior = amd64_linux_child_post_startup_inferior;

  /* Add our register access methods.  */
  t->to_fetch_registers = amd64_linux_fetch_inferior_registers;
  t->to_store_registers = amd64_linux_store_inferior_registers;

  /* This functions make elfcore note sections. */
  linux_elfcore_write_prpsinfo = amd64_linux_elfcore_write_prpsinfo;
  linux_elfcore_write_prstatus = amd64_linux_elfcore_write_prstatus;
  linux_elfcore_write_prfpreg = amd64_linux_elfcore_write_prfpreg;

  /* Register the target.  */
  linux_nat_add_target (t);

  observer_attach_linux_new_thread (amd64_linux_new_thread);
}
